
package hid4Java;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.bluetooth.*;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.StreamConnectionNotifier;

public class BtHidHandler {
    private final BtHidHandler btHandler;
    private Logger log;
    
    private static final int MINORDEVICECLASS = 0x0000C4; // definiert ein keyboard, maus, joystick Hid gerät ->  // id 0x0202
    private static final int MAJORDEVICECLASS = 0x000500; // definiert ein Peripheral device 
    private static final int HIDSERVICECLASS = 0x080000; // Definiert die Service Class: Capturing(HID)
    
    private final short iLangId_enUS = 0x0409;
    private final short iBaseId_enUS = 0x0100; // Offset = 0
    private final short iLangId_deDE = 0x0407;
    private final short iBaseId_deDE = 0x0100; // Offset = 0
    
    private final byte bReportType_Mouse = 0x22; // 0x22 HID // 0x23 PID
    private final byte bReportType_Keyboard = 0x22; // 0x22 HID // 0x23 PID
    private final byte bReportType_Joystick = 0x22; // 0x22 HID // 0x23 PID
    
    private final UUID btProtocolId = new UUID("0100", true); // id 0x0003 -> RFCOMM(btspp) / id 0x0100 -> l2cap
    private final UUID btServiceIdHid = new UUID("1124", true); // Id 0x0001 wird lib-intern gesetzt
    private final UUID btProfileIdHid = new UUID("0011", true);
    private final short btHidProfileVersion = 0x0100; // Version 1.00 ist selbstgewählt
    private final short btPSMHidInterrupt = 0x0013;
    
    private final short btLanguageCode = 0x656E; // code_ISO639
    private final short btLanguageEncoding = 0x006A; // UTF-8
    private final short btLanguageBaseId = 0x0100; // Offset = 0
    
    private final DataElement btHIDServiceName = new DataElement(DataElement.STRING, "MultiDeskBoard Service"); // id 0x0100
    private final DataElement btHIDDeviceReleaseNumber = new DataElement(DataElement.U_INT_2, 0x0100); // id 0x0200, Version 1.00 selfdefined
    private final DataElement btHIDParserVersion = new DataElement(DataElement.U_INT_2, 0x0111); // id 0x0201, must match the USB Report specification version use for the reports
    private final DataElement btHIDDeviceSubclass = new DataElement(DataElement.U_INT_1, MINORDEVICECLASS);  // id 0x0202, classes like in CoD
    private final DataElement btHIDCountryCode = new DataElement(DataElement.U_INT_1, 0);  // id 0x0203, 0 -> generic / no special keyboard language
    private final DataElement btHIDVirtualCable = new DataElement(true); // id 0x0204, true -> single host connection mode
    private final DataElement btHIDReconnectInitiate = new DataElement(true); // id 0x0205, true -> active reconnect tries after connection lost
    private final DataElement btHIDBatteryPower = new DataElement(false); // id 0x0209, false -> not battery power, no energy-saving-automatics active
    private final DataElement btHIDRemoteWake = new DataElement(true); // id 0x020A, true -> device ist registered as "allowed to wake host" in the host
    private final DataElement btHIDProfileVersion = new DataElement(DataElement.U_INT_2, 0x0100); // id 0x020B, shall match the Version of used BT-Hid-Spec
    private final DataElement btHIDNormallyConnectable = new DataElement(true); // id 0x020D, true -> allows the host to (re)astablish the connection without interaction
    private final DataElement btHIDBootDevice = new DataElement(true); // id 0x020E, true -> device supports boot-mode (Hid Reports must contain at least the basic-standard format for keyboard and mice)
    
    private BtHidTaskHandler taskHandler = null;
    private final ExecutorService hidThread = Executors.newCachedThreadPool();
    private Future hidTask = null;
    
    public BtHidHandler(Logger log){
        btHandler = this;
        btHandler.log = log;
    }
    
    // publics
    public final void start() throws RuntimeException,NullPointerException,IllegalArgumentException,IOException {
        if (!LocalDevice.isPowerOn()){throw new BluetoothStateException("Device off");}
        if (hidTask != null){hidTask.cancel(true);}
        //File mouseReport = new File("F:\\Projekt_MultiDeskBoard\\Doku\\Pi_Konfiguration\\PiConfigs\\mediaMouseRepDesc.dat");
        //File keyboardReport = new File("F:\\Projekt_MultiDeskBoard\\Doku\\Pi_Konfiguration\\PiConfigs\\mediaKeyboardRepDesc.dat");
        //File joystickReport = new File("F:\\Projekt_MultiDeskBoard\\Doku\\Pi_Konfiguration\\PiConfigs\\mediaJoystickRepDesc.dat");
        File mouseReport = new File("/home/mdb/MDB_App/PiConfigs/mediaMouseRepDesc.dat");
        File keyboardReport = new File("/home/mdb/MDB_App/PiConfigs/mediaKeyboardRepDesc.dat");
        File joystickReport = new File("/home/mdb/MDB_App/PiConfigs/mediaJoystickRepDesc.dat");
        taskHandler = new BtHidTaskHandler(mouseReport, keyboardReport, joystickReport);
        hidTask = hidThread.submit(taskHandler);
        //LocalDevice.getLocalDevice().setDiscoverable(DiscoveryAgent.LIAC);
    }
    @Override
    public final String toString(){
        try {
            LocalDevice loclDev = LocalDevice.getLocalDevice();
            DeviceClass devCls = loclDev.getDeviceClass();
            return String.format("MAC: %s with CoD: 0x%06X providing %s", 
                    loclDev.getBluetoothAddress(), (devCls.getServiceClasses() | devCls.getMajorDeviceClass() | devCls.getMinorDeviceClass()), devCls);
        } catch (BluetoothStateException ex) {
            return "";
        }
    }
    
    // private
    private String serverUrlBuilder(String sProtocol, String sServer, UUID service, String[] sOptions){
        return sProtocol + ":" + sServer + ":" + service.toString() + ";" + String.join(";", sOptions);
    }
    private ServiceRecord buildHidService(ServiceRecord record, File reportDatFileMouse, File reportDatFileKeyboard, File reportDatFileJoystick) 
            throws IOException,RuntimeException,IllegalArgumentException {
        DataElement btBluetoothProfileDescriptorList = new DataElement(DataElement.DATSEQ);
        btBluetoothProfileDescriptorList.addElement(buildBluetoothProfileDescriptorListEntry(btServiceIdHid, btHidProfileVersion));
        
        DataElement btAdditionalProtocolDescriptorLists = new DataElement(DataElement.DATSEQ);
        btAdditionalProtocolDescriptorLists.addElement(buildAdditionalProtocolDescriptorListEntry(btProtocolId, btPSMHidInterrupt, btProfileIdHid));
        
        DataElement btHIDDescriptorList = new DataElement(DataElement.DATSEQ);
        btHIDDescriptorList.addElement(buildHIDDescriptorListEntry(reportDatFileMouse, bReportType_Mouse));
        btHIDDescriptorList.addElement(buildHIDDescriptorListEntry(reportDatFileKeyboard, bReportType_Keyboard));
        btHIDDescriptorList.addElement(buildHIDDescriptorListEntry(reportDatFileJoystick, bReportType_Joystick));
        
        DataElement btHIDLANGIDBaseList = new DataElement(DataElement.DATSEQ);
        btHIDLANGIDBaseList.addElement(buildHIDLANGIDBaseListEntry(iLangId_enUS, iBaseId_enUS));
        btHIDLANGIDBaseList.addElement(buildHIDLANGIDBaseListEntry(iLangId_deDE, iBaseId_deDE));
        // Attribute 0x0000 "ServiceRecordHandle" is autogenerated at server creation by bluecove
        // Attribute 0x0001 "ServiceClassIDList" is autogenerated at server creation by bluecove
        // Attribute 0x0004 "ProtocolDescriptorList" is autogenerated at server creation by bluecove
        record.setAttributeValue(0x0006, buildLanguageBaseAttributeIDList(btLanguageCode, btLanguageEncoding, btLanguageBaseId));
        record.setAttributeValue(0x0009, btBluetoothProfileDescriptorList);
        record.setAttributeValue(0x000D, btAdditionalProtocolDescriptorLists);
        record.setAttributeValue(0x0100, btHIDServiceName);
        record.setAttributeValue(0x0200, btHIDDeviceReleaseNumber);
        record.setAttributeValue(0x0201, btHIDParserVersion);
        record.setAttributeValue(0x0202, btHIDDeviceSubclass);
        record.setAttributeValue(0x0203, btHIDCountryCode);
        record.setAttributeValue(0x0204, btHIDVirtualCable);
        record.setAttributeValue(0x0205, btHIDReconnectInitiate);
        record.setAttributeValue(0x0206, btHIDDescriptorList);
        record.setAttributeValue(0x0207, btHIDLANGIDBaseList);
        record.setAttributeValue(0x0209, btHIDBatteryPower);
        record.setAttributeValue(0x020A, btHIDRemoteWake);
        record.setAttributeValue(0x020B, btHIDProfileVersion);
        record.setAttributeValue(0x020D, btHIDNormallyConnectable);
        record.setAttributeValue(0x020E, btHIDBootDevice);
        return record;
    }
    private DataElement buildLanguageBaseAttributeIDList(short iCode, short iEncoding, short iBaseId){
        DataElement btLanguageBaseAttributeIDList = new DataElement(DataElement.DATSEQ);
        btLanguageBaseAttributeIDList.addElement(new DataElement(DataElement.U_INT_2, iCode));
        btLanguageBaseAttributeIDList.addElement(new DataElement(DataElement.U_INT_2, iEncoding));
        btLanguageBaseAttributeIDList.addElement(new DataElement(DataElement.U_INT_2, iBaseId));
        return btLanguageBaseAttributeIDList;
    }
    private DataElement buildBluetoothProfileDescriptorListEntry(UUID profile, short iVersion){
        DataElement btBluetoothProfile = new DataElement(DataElement.DATSEQ);
        btBluetoothProfile.addElement(new DataElement(DataElement.UUID, profile));
        btBluetoothProfile.addElement(new DataElement(DataElement.U_INT_2, iVersion));
        return btBluetoothProfile;
    }
    private DataElement buildAdditionalProtocolDescriptorListEntry(UUID protocol, short iPSM, UUID profile){
        DataElement btAdditionalProtocolDescriptor = new DataElement(DataElement.DATSEQ);
        DataElement btDescriptor0 = new DataElement(DataElement.DATSEQ);
        DataElement btDescriptor1 = new DataElement(DataElement.DATSEQ);
        btAdditionalProtocolDescriptor.addElement(btDescriptor0);
        btAdditionalProtocolDescriptor.addElement(btDescriptor1);
        btDescriptor0.addElement(new DataElement(DataElement.UUID, protocol));
        btDescriptor0.addElement(new DataElement(DataElement.U_INT_2, iPSM));
        btDescriptor1.addElement(new DataElement(DataElement.UUID, profile));
        return btAdditionalProtocolDescriptor;
    }
    private DataElement buildHIDLANGIDBaseListEntry(short iLangId, short iBaseId){
        DataElement btHIDLANGIDBase = new DataElement(DataElement.DATSEQ);
        btHIDLANGIDBase.addElement(new DataElement(DataElement.U_INT_2, iLangId));
        btHIDLANGIDBase.addElement(new DataElement(DataElement.U_INT_2, iBaseId));
        return btHIDLANGIDBase;
    }
    private DataElement buildHIDDescriptorListEntry(File reportDatFile, byte bReportType) throws IOException{
        DataElement btHIDDescriptor = new DataElement(DataElement.DATSEQ);
        btHIDDescriptor.addElement(new DataElement(DataElement.U_INT_1, bReportType));
        
        StringBuilder btHIDDescriptor_Data = new StringBuilder();
        ArrayList<Byte> report = HidWorkbench.parseDatFormat(new BufferedReader(new FileReader(reportDatFile)));
        report.forEach(bElement -> {
            btHIDDescriptor_Data.append((char)bElement.byteValue());
        });
        btHIDDescriptor.addElement(new DataElement(DataElement.STRING, btHIDDescriptor_Data.toString()));
        
        return btHIDDescriptor;
    }
    
    // Verbindungsklasse
    private class BtHidTaskHandler implements Runnable {
        private final StreamConnectionNotifier hidNotifier;
        
        public BtHidTaskHandler(File reportDatFileMouse, File reportDatFileKeyboard, File reportDatFileJoystick) 
                throws RuntimeException,NullPointerException,IllegalArgumentException,IOException {
            hidNotifier = (StreamConnectionNotifier)Connector.open(
                    serverUrlBuilder("btspp", "//localhost", btServiceIdHid, new String[]{"authenticate=true","authorize=true","encrypt=true"})
                    , Connector.READ_WRITE, true);
            LocalDevice localDevice = LocalDevice.getLocalDevice();
            localDevice.updateRecord(buildHidService(getService(localDevice), reportDatFileMouse, reportDatFileKeyboard, reportDatFileJoystick));
        }
        
        public final ServiceRecord getService() throws IOException,NullPointerException,IllegalArgumentException{
            return getService(LocalDevice.getLocalDevice());
        }
        public final ServiceRecord getService(LocalDevice localDevice) throws NullPointerException,IllegalArgumentException {
            return localDevice.getRecord(hidNotifier);
        }
        
        @Override
        public void run() {
            try{
                log.log(Level.INFO, "Warte auf Hid Verbindung...");
                StreamConnection hidConnection = (StreamConnection)hidNotifier.acceptAndOpen();
                log.log(Level.INFO, "HID Connection geöffnet");

                // die verbindung testen
                RemoteDevice myRD = RemoteDevice.getRemoteDevice(hidConnection);
                if (myRD.isAuthenticated() && myRD.isAuthorized(hidConnection) && myRD.isEncrypted()){
                    // melde erfolg
                    log.log(Level.INFO, "yippie!");
                } else {
                    // melde grütze
                    log.log(Level.INFO, "mist!");
                }

                try{Thread.sleep(600000);}catch(InterruptedException ex){}

                hidNotifier.close();
                log.log(Level.INFO, "HID Connection geschlossen");
                
                
                
            }catch(NullPointerException | IOException ex){
                log.log(Level.INFO, "Huston we have a runtime problem: {0}", ex.toString());
            }
        }
    }
}
